"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const lodash_1 = require("lodash");
const core_node_1 = require("@cirrusct/core-node");
const mr_build_1 = require("@cirrusct/mr-build");
const DeployConfigResolver_1 = require("./DeployConfigResolver");
const DeployTemplateContext_1 = require("./DeployTemplateContext");
const config_1 = require("./config");
exports.handleGenerateCommand = ({ logger, packages, options = {}, cwd, monoRepo, }) => tslib_1.__awaiter(this, void 0, void 0, function* () {
    try {
        if (options.deployment) {
            const resolvedConfigs = yield config_1.getResolvedConfigs(options.deployment, options, monoRepo);
            yield exports.generate(resolvedConfigs, logger);
        }
        else {
            const packagesWithDeployments = yield config_1.getPackagesWithDeployments(monoRepo, packages);
            for (const pkg of packagesWithDeployments) {
                const childLogger = logger.child({ package: pkg.name });
                childLogger.info('Running Package Deploy');
                yield exports.generatePackageDeployment(pkg, options, logger);
            }
        }
    }
    catch (e) {
        throw core_node_1.createError(`Deploy Command Failed`, e);
    }
});
exports.generatePackageDeployment = (mrPackage, options, logger, name) => tslib_1.__awaiter(this, void 0, void 0, function* () {
    const buildConfig = yield mr_build_1.MrBuild.getPackageBuildConfig(mrPackage);
    if (buildConfig.specs) {
        const mr = mrPackage.monoRepo;
        const deployConfig = mr.config.deploy;
        // TODO: centralize
        const spec = lodash_1.get(buildConfig.specs, 'default', lodash_1.get(buildConfig.specs, Object.keys(buildConfig.specs)[0]));
        const resolver = new DeployConfigResolver_1.DeploymentConfigResolver(mrPackage, deployConfig, spec, options, name);
        const resolved = yield resolver.resolve();
        yield exports.generate(resolved, logger);
    }
    else {
        logger.info({}, `Skipping package ${mrPackage.name}.  No Build Spec`);
    }
});
exports.generate = (resolvedConfig, logger) => tslib_1.__awaiter(this, void 0, void 0, function* () {
    const deploymentConfigs = Array.isArray(resolvedConfig) ? resolvedConfig : [resolvedConfig];
    if (deploymentConfigs && deploymentConfigs.length > 0) {
        logger.info(`Generating ${deploymentConfigs.length} deployment${deploymentConfigs.length === 1 ? '' : 's'}`);
        for (const config of deploymentConfigs) {
            const childLogger = logger.child({ deployment: config.name, output: config.output });
            childLogger.info(`Generating deployment`);
            const serverContexts = config.dependsOnServersResolvedConfigs.map(serverConfig => {
                return new DeployTemplateContext_1.DeployTemplateContext(serverConfig);
            });
            childLogger.debug(`Removing existing output folder`);
            core_node_1.rimraf(config.output);
            const context = new DeployTemplateContext_1.DeployTemplateContext(config, serverContexts);
            for (const templateFile of context.templateFiles) {
                childLogger.debug({ source: templateFile.source.filePath, output: templateFile.outputPath }, `Generating template..'`);
                templateFile.render();
            }
            for (const serverContext of serverContexts) {
                const serverLogger = childLogger.child({ serverPackage: serverContext.mrPackage.name });
                serverLogger.info({ serverPackage: serverContext.mrPackage.name }, `Generating server templates..'`);
                context.currentServerContext = serverContext;
                for (const templateFile of serverContext.templateFiles) {
                    templateFile.context = context;
                    serverLogger.debug({
                        source: templateFile.source.filePath,
                        output: templateFile.getOutputPath(serverContext.output),
                    }, `Generating server template..'`);
                    templateFile.render(serverContext.output);
                }
            }
        }
    }
    else {
        logger.info(`No Deployment Configs Resolved`);
    }
});
//# sourceMappingURL=generate.js.map